// TODO: Add bounding boxes, optimize calls to drawwall
// get rid of sc.scale?
// split this file!

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <pc.h>
#include <kb.h>
#include <bios.h>
#include <sys/movedata.h>
#include "blin.h"
#include "yuck.h"

#define VecList(V) V[0], V[1], V[2]

#define VecLen2(V) sqrt(VecDot2(V, V))
#define VecList2(V) V[0], V[1]
#define VecDot2(A, B) ((A[0]*B[0]) + (A[1]*B[1]))

#define VecScalar2(V, assign_op, k) \
(	V[0] assign_op k, \
	V[1] assign_op k \
)
#define VecVec2(A, assign_op, B) \
(	A[0] assign_op B[0], \
	A[1] assign_op B[1] \
)
#define VecVecVec2(A, assign_op, B, op, C) \
(	A[0] assign_op B[0] op C[0], \
	A[1] assign_op B[1] op C[1] \
)
#define MatVec2(A, M, B) \
(	A[0] = M[0][0]*B[0] + M[0][1]*B[1], \
	A[1] = M[1][0]*B[0] + M[1][1]*B[1] \
)

#define epsilon 1e-10
/*
#define appendwall(a, b) \
	a->next = (wlelem *) xmalloc(sizeof(wlelem)); \
	a = a->next; \
	a->next = NULL; \
	a->w = b;
	*/

#define texel(n, tx, ty) tex[n].data[ty * tex[n].width + tx]

#define angleinc M_PI / 40

#define frontclip 5.0
#define playersize 12.0

#define mapminzoom 0.06
#define mapmaxzoom 1

enum {
	t_back = -1,
	t_front = 1,
	t_on = 0,

	/* t_cross_plus means it crosses, and the wall start is in front,
	 * wall end is behind, and t_cross_minus is the other way around */
	t_cross_plus = 2,
	t_cross_minus = -2,
	string_max = 80,
	SHEIGHT = 200,
	SWIDTH = 320,
	debug = 0
};

typedef double vector[3];
typedef double matrix[3][3];
typedef double vector2[2];
typedef double matrix2[2][2];

struct wallattr {
	//double red, green, blue;
	int texture;
	double offset;
	short visible;
};

typedef struct wallattr wallattr;

struct origwall {
	int start[2];
	int end[2];
	wallattr plus, minus;
};

typedef struct origwall origwall;

struct wall {
	vector2 start;
	vector2 end;
	vector2 n;
	double length;
	wallattr plus, minus;
};

typedef struct wall wall;

struct wlelem {
	struct wlelem *next;
	wall w;
};

typedef struct wlelem wlelem;

struct texture_map {
	int width;
	int height;
	int logheight;
	int wmask;
	int hmask;
	byte *data;
	char filename[string_max];
};

typedef struct texture_map texture_map;

struct plane2 {
	vector2 p;
	vector2 n;
};

typedef struct plane2 plane2;

struct bsptreenode {
	struct bsptreenode *front;
	struct bsptreenode *back;
	plane2 cut;
	wlelem *wl;
};

typedef struct bsptreenode bsptreenode;

struct sceneinfo {
	vector eye;
	vector at;
	vector up;
	double focal;
	int swidth;
	int sheight;
	int smidx, smidy;
	double scale;
	vector2 leftn, rightn;
};

typedef struct sceneinfo sceneinfo;

// prototypes
void screenshot(void);
void clipmove(vector2 src, vector2 dst, bsptreenode *node);
void key_map(void);
void key_other(void);
void key_editwall(void);
void key_movement(void);
wall *splitwall(wall *w, vector2 p);
void orig2wall(origwall *ow, wall *w);
void insertwall(bsptreenode *node, wall *w);
void addwalltofront(bsptreenode *node, wall *w);
void addwalltoback(bsptreenode *node, wall *w);
char classify(wall *w, plane2 *cut);
void savewallattr(FILE *, wallattr *wattr);
void savefile(char *filename);
void showtexture(int x, int y, texture_map *t);
void painter(bsptreenode *node);
void drawwall(wall *w, char);
void drawmap(void);

void inorderdrawmap(bsptreenode *r);
void drawscene(void);
void plot(int x, int y, byte c);
void bresenham(int x1, int y1, int x2, int y2, byte c);
void printwall(wall *);
void preorderprint(bsptreenode *);
void printwalllist(wlelem *);
void makebsptree(bsptreenode *, wlelem *);
void readin(char *, void *, char *, ...);
void parsecomment(void);
void loadtex(int n, char *filename);
void loadfile(char *);

void *normalize2(double [2]);
void readinwallattr(wallattr *);

int wallmax, wallcount;
wlelem *wlhead;
bsptreenode *bsproot;
origwall *orig;

char stopflag;
char mapmode;
char clipmode;

FILE *infile;

sceneinfo sc;

//double angleinc = M_PI / 32;
double sinangleinc, cosangleinc;
double stepsize = 10;

//matrix2 transform;

char img[SWIDTH * SHEIGHT];
texture_map *tex;
int texcount;

int ceilheight = 128;
int floorheight = 0;
int eyeheight = 64;

int editwall;
int offset=0;

double mapzoom;

char oldkey[128];

int main(int argc, char *argv[]) {
	int i;
	int x1, y1, x2, y2;
	wlelem *we;
	/*
	window w;
	drawing d;
	*/

	if (argc < 2) {
		fprintf(stderr, "No filename given\n");
		exit(2);
	}

	loadfile(argv[1]);

	cosangleinc = cos(angleinc);
	sinangleinc = sin(angleinc);

	bsproot = (bsptreenode *) xmalloc(sizeof(bsptreenode));
	makebsptree(bsproot, wlhead);

	/*
	printf("BSP tree:\n");
	preorderprint(bsproot);
	
	printf("End of BSP tree\n");
	printf("%d\n", wallcount);

	we = wlhead->next;

	while(we) {
		printf("%lf %lf %lf %lf\n", we->w.start[0], we->w.start[1],
				we->w.end[0], we->w.end[1]);
		printf("%lf %lf\n", we->w.n[0], we->w.n[1]);
		we = we->next;
	}
	*/

	initgraphics();

	mapzoom = 0.5;
	editwall = -1;
	clipmode = 1;

	sc.swidth = SWIDTH;
	sc.sheight = SHEIGHT;
	sc.smidx = sc.swidth / 2;
	sc.smidy = sc.sheight / 2;

	sc.focal = 200;
	sc.scale = 1;
	sc.rightn[0] = -(sc.leftn[0] = sc.focal);
	sc.rightn[1] = sc.leftn[1] = sc.swidth / 2.0;

	logprintf("focal: %lf", sc.focal);

	mapmode = 1;
	stopflag = 0;
	kb_install(0);

	// Main Loop
	while(!stopflag) {
		drawscene();
		kb_update();
		if (mapmode) {
			key_map();
		}
		if (editwall >= 0) {
			key_editwall();
		}
		key_movement();
		key_other();
		for (i=0; i<128; i++) {
			oldkey[i] = kb_key(i);
		}
	}

	settextmode();
}

void key_map(void) {
	if (kb_key(KB_SCAN_MINUS_PAD)) {
		if (mapzoom > mapminzoom) mapzoom-=0.01;
	}
	if (kb_key(KB_SCAN_PLUS_PAD)) {
		if (mapzoom < mapmaxzoom) mapzoom+=0.01;
	}
}

void key_other(void) {
	if (kb_key(KB_SCAN_TAB) && !oldkey[KB_SCAN_TAB]) {
		mapmode = !mapmode;
	}

	if (kb_key(KB_SCAN_S)) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			savefile("temp");
		}
		/*
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			sc.scale /= 2;
		} else {
			sc.scale *= 2;
		}
		*/
	}
	if (kb_key(KB_SCAN_E)) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			eyeheight-=stepsize;
		} else {
			eyeheight+=stepsize;
		}
	}

	if (kb_key(KB_SCAN_F) && !oldkey[KB_SCAN_F]) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			sc.focal /= 2;
			sc.rightn[0] = -(sc.leftn[0] = sc.focal);
		} else {
			sc.focal *= 2;
			sc.rightn[0] = -(sc.leftn[0] = sc.focal);
		}
	}

	if (kb_key(KB_SCAN_C) && !oldkey[KB_SCAN_C]) {
		clipmode = !clipmode;
	}

	if (kb_key(KB_SCAN_R) && !oldkey[KB_SCAN_R]) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			if (ceilheight > 1) {
				ceilheight /= 2;
				eyeheight /= 2;
			}
		} else {
			ceilheight *= 2;
			eyeheight *= 2;
		}
	}

	if (kb_key(KB_SCAN_N) && editwall == -1) {
		vector2 v;

		editwall = wallcount;
		wallcount++;

		if (wallcount > wallmax) {
			origwall *temp;

			wallmax *= 2;
			temp = orig;
			orig = (origwall *) xmalloc(sizeof(origwall) * wallmax);
			memcpy(orig, temp, sizeof(origwall) * (wallcount - 1));
			free(temp);
		}

		v[0] = -sc.at[1];
		v[1] = sc.at[0];
		VecVecVec2(orig[editwall].start, =, sc.eye, +, 64 * sc.at);
		VecVec2(orig[editwall].end, =, orig[editwall].start);
		VecVec2(orig[editwall].start, +=, 64 * v);
		VecVec2(orig[editwall].end, -=, 64 * v);
		orig[editwall].plus = orig[0].plus;
		orig[editwall].minus = orig[0].minus;
	}

	if (kb_key(KB_SCAN_Q)) {
		stopflag = 1;
	}

	if (kb_key(KB_SCAN_F10)) {
		screenshot();
	}
}

void key_editwall(void) {
	vector2 v;

	if (kb_key(KB_SCAN_O)) {
		offset--;
	}

	if (kb_key(KB_SCAN_P)) {
		offset++;
	}

	if (kb_key(KB_SCAN_PERIOD)) {
		if (kb_shift() & KB_SHIFT_ANY_CONTROL) {
			orig[editwall].minus.texture++;
			orig[editwall].minus.texture %= texcount;
		} else {
			orig[editwall].plus.texture++;
			orig[editwall].plus.texture %= texcount;
		}
	}

	if (kb_key(KB_SCAN_COMMA)) {
		if (kb_shift() & KB_SHIFT_ANY_CONTROL) {
			orig[editwall].minus.texture += texcount - 1;
			orig[editwall].minus.texture %= texcount;
		} else {
			orig[editwall].plus.texture += texcount - 1;
			orig[editwall].plus.texture %= texcount;
		}
	}

	if (kb_key(KB_SCAN_ESC)) {
		wall tempwall;

		orig2wall(&orig[editwall], &tempwall);
		insertwall(bsproot, &tempwall);
		editwall = -1;
	}

	if (kb_key(KB_SCAN_4_PAD)) {
		v[0] = -sc.at[1];
		v[1] = sc.at[0];
		if (kb_shift() & KB_SHIFT_ANY_CONTROL) {
			VecVec2(orig[editwall].end, +=, stepsize * v);
		} else {
			VecVec2(orig[editwall].start, +=, stepsize * v);
		}
	}
	if (kb_key(KB_SCAN_6_PAD)) {
		v[0] = sc.at[1];
		v[1] = -sc.at[0];
		if (kb_shift() & KB_SHIFT_ANY_CONTROL) {
			VecVec2(orig[editwall].end, +=, stepsize * v);
		} else {
			VecVec2(orig[editwall].start, +=, stepsize * v);
		}
	}
	if (kb_key(KB_SCAN_8_PAD)) {
		if (kb_shift() & KB_SHIFT_ANY_CONTROL) {
			VecVec2(orig[editwall].end, +=, stepsize * sc.at);
		} else {
			VecVec2(orig[editwall].start, +=, stepsize * sc.at);
		}
	}
	if (kb_key(KB_SCAN_2_PAD)) {
		if (kb_shift() & KB_SHIFT_ANY_CONTROL) {
			VecVec2(orig[editwall].end, -=, stepsize * sc.at);
		} else {
			VecVec2(orig[editwall].start, -=, stepsize * sc.at);
		}
	}
}

void key_movement(void) {
	vector2 v;
	double x, y;
	vector2 dst;

	VecVec2(dst, =, sc.eye);

	if ((kb_key(KB_SCAN_4_PAD) && editwall == -1) ||
			kb_key(KB_SCAN_LEFT)) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			v[0] = -sc.at[1];
			v[1] = sc.at[0];
			VecVec2(dst, +=, stepsize * v);
		} else {
			x = sc.at[0];
			y = sc.at[1];
			sc.at[0] = cosangleinc * x - sinangleinc * y;
			sc.at[1] = sinangleinc * x + cosangleinc * y;
		}
	}
	if ((kb_key(KB_SCAN_6_PAD) && editwall == -1) ||
			kb_key(KB_SCAN_RIGHT)) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			v[0] = sc.at[1];
			v[1] = -sc.at[0];
			VecVec2(dst, +=, stepsize * v);
		} else {
			x = sc.at[0];
			y = sc.at[1];
			sc.at[0] = cosangleinc * x + sinangleinc * y;
			sc.at[1] = -sinangleinc * x + cosangleinc * y;
		}
	}
	if ((kb_key(KB_SCAN_8_PAD) && editwall == -1) ||
			kb_key(KB_SCAN_UP)) {
		VecVec2(dst, +=, stepsize * sc.at);
	}
	if ((kb_key(KB_SCAN_2_PAD) && editwall == -1) ||
			kb_key(KB_SCAN_DOWN)) {
		VecVec2(dst, -=, stepsize * sc.at);
	}

	if (clipmode) {
		clipmove(sc.eye, dst, bsproot);
	} else {
		VecVec2(sc.eye, =, dst);
	}
}

void clipmove(vector2 src, vector2 dst, bsptreenode *node) {
	wlelem *we;
	vector2 p, p2, v, dir;
	double dsrc, ddst, mu;
	char type;
	char flag;

	VecVecVec2(p, =, src, -, node->cut.p);
	dsrc = VecDot2(node->cut.n, p);
	VecVecVec2(p, =, dst, -, node->cut.p);
	ddst = VecDot2(node->cut.n, p);

	flag = 0;
	/*
	VecVecVec2(dir, =, dst, -, src);
	VecVecVec2(v, =, node->cut.p, -, src);
	mu = VecDot2(node->cut.n, dir);
	if (fabs(mu) > epsilon) {
		mu = VecDot2(node->cut.n, v) / VecDot2(node->cut.n, dir);
		VecVecVec2(p, =, src, +, mu*dir);

		we = node->wl;
		while (we->next && !flag) {
			we = we->next;
			// Is there actually a wall in the way?
			if ((we->w.start[1] < p[1] && p[1] < we->w.end[1]) ||
					(we->w.start[1] > p[1] && p[1] > we->w.end[1])) {
				flag = 1;
			}
			if ((we->w.start[0] < p[0] && p[0] < we->w.end[0]) ||
					(we->w.start[0] > p[0] && p[0] > we->w.end[0])) {
				flag = 1;
			}
		}
	}
	*/
	VecVecVec2(p, =, src, -, dsrc*node->cut.n);
	VecVecVec2(p2, =, dst, -, ddst*node->cut.n);
		we = node->wl;
		while (we->next && !flag) {
			we = we->next;
			// Is there actually a wall in the way?
			if ((we->w.start[1] < p[1] && p[1] < we->w.end[1]) ||
					(we->w.start[1] > p[1] && p[1] > we->w.end[1])) {
				flag = 1;
			}
			if ((we->w.start[0] < p[0] && p[0] < we->w.end[0]) ||
					(we->w.start[0] > p[0] && p[0] > we->w.end[0])) {
				flag = 1;
			}
			if ((we->w.start[0] < p2[0] && p2[0] < we->w.end[0]) ||
					(we->w.start[0] > p2[0] && p2[0] > we->w.end[0])) {
				flag = 1;
			}
			if ((we->w.start[0] < p2[0] && p2[0] < we->w.end[0]) ||
					(we->w.start[0] > p2[0] && p2[0] > we->w.end[0])) {
				flag = 1;
			}
		}

	if (dsrc > epsilon) {
		if (flag && (ddst < playersize)) {
			//VecVecVec2(dst, =, p, +, playersize * node->cut.n);
			mu = (playersize - ddst);
			mu = mu > dsrc ? dsrc : mu;
			VecVec2(dst, +=, mu * node->cut.n);
		}
		if (node->front) {
			clipmove(src, dst, node->front);
		} else {
			VecVec2(src, =, dst);
		}
	} else if (dsrc < -epsilon) {
		dsrc = -dsrc;
		if (flag && (-ddst < playersize)) {
			//VecVecVec2(dst, =, p, -, playersize * node->cut.n);
			mu = (playersize + ddst);
			mu = mu > dsrc ? dsrc : mu;
			VecVec2(dst, -=, mu * node->cut.n);
		}
		if (node->back) {
			clipmove(src, dst, node->back);
		} else {
			VecVec2(src, =, dst);
		}
	} else {
		// Should not be here (can't be on a wall!)
		//VecVec2(src, =, dst);
	}

}

void drawscene(void) {
	int i, j;
	vector2 p;
	wall tempwall;
	double d;

	for (i=0; i<(sc.swidth * sc.sheight); i++) img[i] = 0;

	painter(bsproot);
	if (editwall >= 0) {
		orig2wall(&orig[editwall], &tempwall);
		VecVecVec2(p, =, sc.eye, -, tempwall.start);
		d = VecDot2(tempwall.n, p);
		tempwall.minus.offset = tempwall.plus.offset = offset;
		if (d > epsilon) {
			drawwall(&tempwall, 2);
		} else if (d < -epsilon) {
			drawwall(&tempwall, -2);
		}
	}
	if (mapmode) {
		drawmap();
	}

	drawimage(img);
}

void painter(bsptreenode *node) {
	wlelem *we;
	vector2 p;
	double d;

	VecVecVec2(p, =, sc.eye, -, node->cut.p);
	d = VecDot2(node->cut.n, p);

	if (d > epsilon) {
		if (node->back) {
			painter(node->back);
		}

		we = node->wl->next;
		while(we) {
			if (we->w.plus.visible) {
				drawwall(&we->w, 1);
			}
			we = we->next;
		}

		if (node->front) {
			painter(node->front);
		}
	} else if (d < epsilon) {
		if (node->front) {
			painter(node->front);
		}
		we = node->wl->next;
		while(we) {
			if (we->w.minus.visible) {
				drawwall(&we->w, -1);
			}
			we = we->next;
		}
		if (node->back) {
			painter(node->back);
		}
	}
}

void drawwall(wall *w, char side) {
	int x1, x2;
	int y1t, y1b, y2t, y2b;
	int dx, dy;
	int dyt, dyb;
	int errort, errorb;
	int yinct, yincb;
	int i, j;
	wallattr *wa;
	matrix2 m;
	vector2 v, v1, v2;
	vector2 nv1, nv2;
	vector2 dir;
	double mu1, mu2;
	double d;
	double recip1, recip2, drecip;
	int ymin, ymax;
	double xmult;
	int tx, ty;
	int dty;
	int hmask, wmask;
	byte *texdata;
	int texlogheight;
	double txstart;
	int temp;

	m[0][0] = sc.scale * sc.at[1];
	m[0][1] = sc.scale * -sc.at[0];
	m[1][0] = sc.scale * sc.at[0];
	m[1][1] = sc.scale * sc.at[1];

	v[0] = w->start[0] - sc.eye[0];
	v[1] = w->start[1] - sc.eye[1];
	MatVec2(v1, m, v);

	v[0] = w->end[0] - sc.eye[0];
	v[1] = w->end[1] - sc.eye[1];
	MatVec2(v2, m, v);

	logprintf("\nTesting points...\n");

	VecVecVec2(dir, =, v2, -, v1);

	if(0) {
		//WRONG! Forgot to transform w->n
	/* v is the normal rotated 90 degrees */
	v[0] = -w->n[1] * side;
	v[1] = w->n[0] * side;
	/* if v and dir face different ways then swap v1 and v2 */
	if (VecDot2(v, dir) < -epsilon) {
		VecVec2(v, =, v1);
		VecVec2(v1, =, v2);
		VecVec2(v2, =, v);
	}
	}

	logprintf("rotated normal: %lf, %lf\n", v[0], v[1]);

	if (side < 0) {
		VecVec2(v, =, v1);
		VecVec2(v1, =, v2);
		VecVec2(v2, =, v);
		VecVec2(dir, =, -dir);
	}
	logprintf("at: %lf, %lf, side: %d\n", sc.at[0], sc.at[1], side);
	logprintf("%lf, %lf -> %lf, %lf\n", v1[0], v1[1], v2[0], v2[1]);
	
	d = VecDot2(sc.leftn, dir);
	VecVec2(nv1, =, v1);
	if (fabs(d) > epsilon) {
		mu1 = -VecDot2(sc.leftn, v1) / d;
		logprintf("mu: %lf\n", mu1);
		if (mu1 > 0 && mu1 <= 1 && (d = v1[1] + mu1*dir[1]) > 0) {
			nv1[0] = v1[0] + mu1*dir[0];
			nv1[1] = d;
			logprintf("Left Intersection %lf, %lf\n", VecList2(nv1));
			logprintf("should be 0: %lf\n", VecDot2(sc.leftn, nv1));
		}
	}

	d = VecDot2(sc.rightn, dir);
	VecVec2(nv2, =, v2);
	if (fabs(d) > epsilon) {
		mu2 = -VecDot2(sc.rightn, v1) / d;
		logprintf("mu: %lf\n", mu2);
		if (mu2 > 0 && mu2 <= 1 && (d = v1[1] + mu2*dir[1]) > 0) {
			nv2[0] = v1[0] + mu2*dir[0];
			nv2[1] = d;
			logprintf("Right Intersection %lf, %lf\n", VecList2(nv2));
			logprintf("should be 0: %lf\n", VecDot2(sc.rightn, nv2));
		}
	}

	logprintf("clipped: %lf, %lf -> %lf, %lf\n", nv1[0], nv1[1], nv2[0], nv2[1]);
	/* if either point is outside, don't draw it */
	if (nv1[1] < frontclip || nv1[1]+(sc.leftn[0]/sc.leftn[1])*nv1[0] < -epsilon
			|| nv1[1]+(sc.rightn[0]/sc.rightn[1])*nv1[0] < -epsilon) {
		logprintf("nv1 outside\n");
		return;
	}
	if (nv2[1] < frontclip || nv2[1]+(sc.leftn[0]/sc.leftn[1])*nv2[0] < -epsilon
			|| nv2[1]+(sc.rightn[0]/sc.rightn[1])*nv2[0] < -epsilon) {
		logprintf("nv2 outside\n");
		return;
	}

	if (side > 0) {
		wa = &w->plus;
	} else {
		wa = &w->minus;
	}

	wmask = tex[wa->texture].wmask;
	hmask = tex[wa->texture].hmask;
	texdata = tex[wa->texture].data;
	texlogheight = tex[wa->texture].logheight;

	logprintf("at: %lf, %lf\n", sc.at[0], sc.at[1]);

	x1 = floor(sc.focal * nv1[0] / nv1[1]) + sc.swidth / 2;
	if (x1 < 0) x1 = 0;

	x2 = (int) (sc.focal * nv2[0] / nv2[1]) + sc.swidth / 2;
	if (x2 >= sc.swidth) x2 = sc.swidth - 1;

	if (x1 > x2) printf("BUG!\n");
	if (x1 == x2) return;
	logprintf("%d %d (x)\n", x1, x2);
	y1t = sc.sheight / 2 - (int) (sc.focal * (ceilheight - eyeheight) / nv1[1]);
	y1b = sc.sheight / 2 - (int) (sc.focal * (floorheight - eyeheight) / nv1[1]);
	y2t = sc.sheight / 2 - (int) (sc.focal * (ceilheight - eyeheight) / nv2[1]);
	y2b = sc.sheight / 2 - (int) (sc.focal * (floorheight - eyeheight) / nv2[1]);

	dx = x2 - x1;
	//d = fabs(v2[1] - v1[1]);
	d = v2[1] - v1[1];

	if (fabs(d) < epsilon) {
		//special, easy case
		xmult = (nv2[0] - nv1[0]) / dx;
		dy = y1b - y1t;
		if (!dy) return;
		ymin = (y1t < 0) ? 0 : y1t;
		ymax = (y1b >= sc.sheight) ? sc.sheight - 1 : y1b;
		temp = (ymax - ymin) * (ceilheight - floorheight) / dy;
		errorb = -(ymax - ymin) / 2;
		for (i=x1; i<=x2; i++) {
			tx = (int) (xmult * (i - x1) + nv1[0] - v1[0]) & wmask;
			ty = ((ceilheight - floorheight) * (ymin - y1t) / dy) & hmask;
			errort = errorb;
			for (j=ymin; j<=ymax; j++) {
				//ty = ((ceilheight - floorheight) * (j - y1t) / dy) & hmask;
				plot(i, j, texdata[(ty << texlogheight) + tx]);
				errort += temp;
				while (errort > 0) {
					ty++;
					ty&=hmask;
					errort -= (ymax - ymin);
				}
			}

		}
	} else {

		/* optimize by unrolling these cases? */
		// (= 4x more code)
		dyt = y2t - y1t;
		if (dyt < 0) {
			yinct = -1;
			dyt = -dyt;
		} else {
			yinct = 1;
		}
		dyb = y2b - y1b;
		if (dyb < 0) {
			yincb = -1;
			dyb = -dyb;
		} else {
			yincb = 1;
		}

		recip1 = 1/nv1[1];
		recip2 = 1/nv2[1];

		drecip = (recip2 - recip1) / dx;

		errorb = errort = -dx / 2;

		xmult = sc.scale * w->length / d;
		dy = y1b - y1t;
		ymin = (y1t < 0) ? 0 : y1t;
		ymax = (y1b >= sc.sheight) ? sc.sheight - 1 : y1b;
		txstart = (v1[1] * xmult) - wa->offset;
		for (i=x1; i<=x2; i++) {

			if (dy) {
				temp =  ((ceilheight - floorheight) << 13) / dy;
				tx = ((int) (xmult / recip1 - txstart)) & wmask;
				for (dty=ymin-y1t, j=i+ymin*sc.swidth; dty<ymax-y1t;
						dty++, j+=sc.swidth) {
					// INNER LOOP - must avoid /, try to avoid *
					//ty = ((ceilheight - floorheight) * dty / dy) & hmask;
					ty = ((temp * dty) >> 13) & hmask;
					img[j] = texdata[tx + (ty << texlogheight)];
				}
			}

			recip1 += drecip;
			errort += dyt;
			while (errort > 0 && y1t != y2t) {
				y1t += yinct;
				dy -= yinct;
				errort -= dx;
				ymin = (y1t < 0) ? 0 : y1t;
			}
				
			errorb += dyb;
			while (errorb > 0 && y1b != y2b) {
				y1b += yincb;
				dy += yincb;
				errorb -= dx;
				ymax = (y1b > sc.sheight) ? sc.sheight - 1 : y1b;
			}
		}
	}

	// outline wall
	// doesn't work because y1t, y1b are destroyed above
	if (0 && abs(side) == 2) {
		bresenham(x1, y1t, x2, y2t, 255);
		bresenham(x1, y1b, x2, y2b, 255);
		bresenham(x1, y1t, x1, y1b, 255);
		bresenham(x2, y2t, x2, y2b, 255);
	}

}

void drawmap(void) {
	byte c;
	int x1, y1, x2, y2;
	int i;
	
	c = 255;

	i = sc.at[0] * playersize * mapzoom;
	x1 = sc.smidx - i;
	x2 = sc.smidx + i;

	i = sc.at[1] * playersize * mapzoom;
	y1 = sc.smidy + i;
	y2 = sc.smidy - i;

	bresenham(x1, y1, x2, y2, c);
	//inorderdrawmap(bsproot);
	for (i=0; i<wallcount; i++) {
		x1 = sc.smidx + (orig[i].start[0] - sc.eye[0]) * mapzoom;
		x2 = sc.smidx + (orig[i].end[0] - sc.eye[0]) * mapzoom;
		y1 = sc.smidy - (orig[i].start[1] - sc.eye[1]) * mapzoom;
		y2 = sc.smidy - (orig[i].end[1] - sc.eye[1]) * mapzoom;
		bresenham(x1, y1, x2, y2, 0xE0);
	}
}

void inorderdrawmap(bsptreenode *r) {
	wlelem *we;
	byte c;
	int x1, x2, y1, y2;
	
	if (r->front) {
		inorderdrawmap(r->front);
	}
	if (r->back) {
		inorderdrawmap(r->back);
	}

	c = 15;

	we = r->wl->next;
	while (we) {
		x1 = (int) we->w.start[0];
		y1 = sc.sheight - (int) we->w.start[1];
		x2 = (int) we->w.end[0];
		y2 = sc.sheight - (int) we->w.end[1];
		bresenham(x1, y1, x2, y2, c);
		x1 = (x1 + x2) / 2;
		y1 = (y1 + y2) / 2;
		if (we->w.plus.visible)
			bresenham(x1, y1, x1 + (int) we->w.n[0] * 10,
					y1 - (int) we->w.n[1] * 10, we->w.plus.texture);
		if (we->w.minus.visible)
			bresenham(x1, y1, x1 - (int) we->w.n[0] * 10,
					y1 + (int) we->w.n[1] * 10, we->w.minus.texture);
		we = we->next;
	}
}

void savefile(char *filename) {
	int i, j, k;
	wlelem *we;
	wall *wp;
	vector2 d;
	FILE *fp;

	if (!(fp = fopen(filename, "w"))) {
			perror(filename);
			exit(1);
	}

	fprintf(fp, "// scene file\n");
	fprintf(fp, "\n// number of textures\n%d", texcount);
	fprintf(fp, "\n// texture filenames:\n");
	for (i=0; i<texcount; i++) {
		fprintf(fp, "%s\n", tex[i].filename);
	}

	fprintf(fp, "\n// number of walls\n%d\n\n", wallcount);

	for (i=0; i<wallcount; i++) {
		fprintf(fp, "\n// wall %d\n", i);
		fprintf(fp, "%d ", orig[i].start[0]);
		fprintf(fp, "%d\n", orig[i].start[1]);
		fprintf(fp, "%d ", orig[i].end[0]);
		fprintf(fp, "%d\n", orig[i].end[1]);
		savewallattr(fp, &orig[i].plus);
		savewallattr(fp, &orig[i].minus);
	}

	fprintf(fp, "\n// player start\n");
	fprintf(fp, "%.0lf ", sc.eye[0]);
	fprintf(fp, "%.0lf\n", sc.eye[1]);
}

void savewallattr(FILE *fp, wallattr *wattr) {
	int i;

	fprintf(fp, "%hd %s\n", wattr->visible, "// visibility flag");

	if (wattr->visible) {
		fprintf(fp, "%d // %s\n", wattr->texture, tex[wattr->texture].filename);
	}
}

void loadfile(char *filename) {
	int i, j, k;
	wlelem *we;
	wall *wp;
	vector2 d;
	FILE *fp;

	if (!(infile = fopen(filename, "r"))) {
			perror(filename);
			exit(1);
	}

	readin("%d", &texcount, "number of textures");
	for (i=0; i<texcount; i++) {
		readin("%s", tex[i].filename, "texture filename");
		fp = infile;
		loadtex(i, tex[i].filename);
		infile = fp;
	}

	readin("%d", &wallcount, "number of walls");

	wallmax = wallcount * 2;
	orig = (origwall *) xmalloc(sizeof(origwall) * wallmax);

	wlhead = (wlelem *) xmalloc(sizeof(wlelem));
	wlhead->next = NULL;

	we = wlhead;

	for (i=0; i<wallcount; i++) {
		we->next = (wlelem *) xmalloc(sizeof(wlelem));
		we = we->next;
		we->next = NULL;
		wp = &we->w;
		readin("%d", &orig[i].start[0], "wall #%d start x", i);
		readin("%d", &orig[i].start[1], "wall #%d start y", i);
		readin("%d", &orig[i].end[0], "wall #%d end x", i);
		readin("%d", &orig[i].end[1], "wall #%d end y", i);

		wp->start[0] = (double) orig[i].start[0];
		wp->start[1] = (double) orig[i].start[1];
		wp->end[0] = (double) orig[i].end[0];
		wp->end[1] = (double) orig[i].end[1];

		wp->n[0] = wp->end[1] - wp->start[1];
		wp->n[1] = wp->start[0] - wp->end[0];
		normalize2(wp->n);

		VecVecVec2(d, =, wp->end, -, wp->start);
		wp->length = sqrt(VecDot2(d, d));
		readinwallattr(&wp->plus);
		orig[i].plus = wp->plus;
		readinwallattr(&wp->minus);
		orig[i].minus = wp->minus;
	}

	readin("%lf", &sc.eye[0], "player start x");
	readin("%lf", &sc.eye[1], "player start y");
	sc.eye[2] = eyeheight;
	sc.at[0] = 1;
	sc.at[1] = 0;
}

void readinwallattr(wallattr *wattr) {
	int i;

	readin("%hd", &wattr->visible, "visibility flag");

	if (wattr->visible) {
		readin("%d", &wattr->texture, "wall #%d texture", i);
	}
	wattr->offset = 0;
}

void readin(char *s, void *f, char *format, ...) {
	va_list arg;
	char done;
	int check;
	char ch;

	done = 0;
	do {
		if (feof(infile)) {
			va_start(arg, format);
			fprintf(stderr, "Unexpected end of file.\nError reading ");
			vfprintf(stderr, format, arg);
			fprintf(stderr, "\n");
			va_end(arg);
			exit(1);
		}

		fscanf(infile, " %c", &ch);
		if (ch == '/') {
			parsecomment();
		} else {
			ungetc(ch, infile);
			check = fscanf(infile, s, f);
			if (!check) {
				va_start(arg, format);
				fprintf(stderr, "Error reading ");
				vfprintf(stderr, format, arg);
				fprintf(stderr, "\n");
				va_end(arg);
				exit(1);
			}
			if (debug) {
				va_start(arg, format);
				fprintf(stderr, "Read ");
				vfprintf(stderr, format, arg);
				fprintf(stderr, "\n");
				va_end(arg);
			}
			done++;
		}
	} while (!done);
}

void parsecomment(void) {
	char ch;
	char flag;

	ch = getc(infile);
	if (ch == '/') {
		while ((ch = getc(infile) != '\n'));
		return;
	} else if (ch != '*') {
		fprintf(stderr, "Warning: bad comment\n", ch);
	}
	flag = 0;
	do {
		do {
			ch = getc(infile);
		} while (ch != '*');
		if ((ch = getc(infile)) == '/') {
			flag++;
		}
	} while (!flag);
}

void *normalize2(double v[2]) {
	double n;

	n = VecLen2(v); //sqrt(v[0] * v[0] + v[1] * v[1]);
	VecScalar2(v, /=, n);
	/*
	v[0] /= n;
	v[1] /= n;
	*/
}

void makebsptree(bsptreenode *node, wlelem *wl) {
	wlelem *wlfronthead, *wlbackhead;
	wlelem *wefront, *weback, *weon, *we;
	wlelem *newwe;
	double d, mu;
	vector2 p, dir, v;
	char type;

	wefront = wlfronthead = (wlelem *) xmalloc(sizeof(wlelem));
	weback = wlbackhead = (wlelem *) xmalloc(sizeof(wlelem));
	wefront->next = weback->next = NULL;
	node->wl = (wlelem *) xmalloc(sizeof(wlelem));
	node->wl->next = NULL;

	weon = node->wl;

	we = wl->next;

	VecVec2(node->cut.p, =, we->w.start);
	VecVec2(node->cut.n, =, we->w.n);

	weon->next = we;
	weon = we;
	we = we->next;
	weon->next = NULL;

	while(we) {
		type = classify(&we->w, &node->cut);
		switch(type) {
			case t_on:
				weon->next = we;
				weon = we;
				we = we->next;
				weon->next = NULL;
				break;

			case t_front:
				wefront->next = we;
				wefront = we;
				we = we->next;
				wefront->next = NULL;
				break;
				
			case t_back:
				weback->next = we;
				weback = we;
				we = we->next;
				weback->next = NULL;
				break;

			case t_cross_plus:
			case t_cross_minus:
				// cut the wall
				// intersection of 2 lines
				VecVecVec2(dir, =, we->w.end, -, we->w.start);
				VecVecVec2(v, =, node->cut.p, -, we->w.start);
				mu = VecDot2(node->cut.n, v) / VecDot2(node->cut.n, dir);
				VecVecVec2(p, =, we->w.start, +, mu*dir);

				VecVecVec2(v, =, node->cut.p, -, p);
				logprintf("Intersection %lf, %lf\n", VecList2(p));
				logprintf("should be 0: %lf\n", VecDot2(node->cut.n, v));
				
				// split the wall
				newwe = (wlelem *) xmalloc(sizeof(wlelem));
				{
					// bad data structure - wlelem should have a
					// wall *, not a wall
					wall *temp;

					//newwe->w = *splitwall(&we->w, p);
					temp = splitwall(&we->w, p);
					newwe->w = *temp;
					xfree(temp);
				}

				// add one wall to the front list, the other to the back
				if (type == t_cross_plus) {
					wefront->next = we;
					wefront = we;
					we = we->next;
					wefront->next = NULL;
					weback->next = newwe;
					weback = newwe;
					weback->next = NULL;
				} else {
					weback->next = we;
					weback = we;
					we = we->next;
					weback->next = NULL;
					wefront->next = newwe;
					wefront = newwe;
					wefront->next = NULL;
				}
				break;
		}
	
	}

	// recurse on front and back nodes
	if (wlfronthead->next) {
		node->front = (bsptreenode *) xmalloc(sizeof(bsptreenode));
		makebsptree(node->front, wlfronthead);
	} else {
		node->front = NULL;
	}
	if (wlbackhead->next) {
		node->back = (bsptreenode *) xmalloc(sizeof(bsptreenode));
		makebsptree(node->back, wlbackhead);
	} else {
		node->back = NULL;
	}
}

char classify(wall *w, plane2 *cut) {
	char type;
	double d, d2;
	vector2 p;

	VecVecVec2(p, =, w->start, -, cut->p);
	d = VecDot2(cut->n, p);
	VecVecVec2(p, =, w->end, -, cut->p);
	d2 = VecDot2(cut->n, p);

	if (d > epsilon) {
		if (d2 >= -epsilon) type = t_front;
		else type = t_cross_plus;
	} else if (fabs(d) < epsilon) {
		if (d2 > epsilon) type = t_front;
		else if (d2 < -epsilon) type = t_back;
		else type = t_on;
	} else {
		if (d2 <= epsilon) type = t_back;
		else type = t_cross_minus;
	}

	return type;
}

void insertwall(bsptreenode *node, wall *w) {
	wlelem *we, *newwe;
	wall *w2;
	double d, mu;
	vector2 p, dir, v;
	char type;

	type = classify(w, &node->cut);
	switch(type) {
		case t_on:
			// append to this node
			we = node->wl;
			while (we->next) we = we->next;
			newwe = (wlelem *) xmalloc(sizeof(wlelem));
			newwe->w = *w;
			newwe->next = NULL;
			we->next = newwe;
			break;

		case t_front:
			addwalltofront(node, w);
			break;
			
		case t_back:
			addwalltoback(node, w);
			break;

		case t_cross_plus:
		case t_cross_minus:
			// cut the wall
			// intersection of 2 lines
			VecVecVec2(dir, =, w->end, -, w->start);
			VecVecVec2(v, =, node->cut.p, -, w->start);
			mu = VecDot2(node->cut.n, v) / VecDot2(node->cut.n, dir);
			VecVecVec2(p, =, w->start, +, mu*dir);
			
			// split the wall
			w2 = splitwall(w, p);

			// add one wall to the front, the other to the back
			if (type == t_cross_plus) {
				addwalltofront(node, w);
				addwalltoback(node, w2);
			} else {
				addwalltofront(node, w2);
				addwalltoback(node, w);
			}
			break;
	}
}

// splits w at point p; w is shortened, and a new wall
// starting from p is returned
wall *splitwall(wall *w, vector2 p) {
	wall *res;

	// clone the wall
	res = (wall *) xmalloc(sizeof(wall));
	*res = *w;

	// new wall's start = intersection = old wall's end
	VecVec2(res->start, =, p);
	VecVec2(w->end, =, p);

	// recalculate wall lengths
	VecVecVec2(p, =, w->end, -, w->start);
	w->length = sqrt(VecDot2(p, p));
	res->length -= w->length;

	// calculate texture offsets
	res->plus.offset += w->length;
	w->minus.offset += res->length;
	return(res);
}

void addwalltofront(bsptreenode *node, wall *w) {
	wlelem *we;

	if (node->front) {
		insertwall(node->front, w);
	} else {
		// make new node
		node->front = (bsptreenode *) xmalloc(sizeof(bsptreenode));
		node = node->front;
		node->front = NULL;
		node->back = NULL;
		node->wl = (wlelem *) xmalloc(sizeof(wlelem));
		VecVec2(node->cut.p, =, w->start);
		VecVec2(node->cut.n, =, w->n);
		we = (wlelem *) xmalloc(sizeof(wlelem));
		we->w = *w;
		we->next = NULL;
		node->wl->next = we;
	}
}

void addwalltoback(bsptreenode *node, wall *w) {
	wlelem *we;

	if (node->back) {
		insertwall(node->back, w);
	} else {
		// make new node
		node->back = (bsptreenode *) xmalloc(sizeof(bsptreenode));
		node = node->back;
		node->front = NULL;
		node->back = NULL;
		node->wl = (wlelem *) xmalloc(sizeof(wlelem));
		VecVec2(node->cut.p, =, w->start);
		VecVec2(node->cut.n, =, w->n);
		we = (wlelem *) xmalloc(sizeof(wlelem));
		we->w = *w;
		we->next = NULL;
		node->wl->next = we;
	}
}

void preorderprint(bsptreenode *n) {
	static int lev = 0;
	int i;

	assert(n);
	
	printf("this: ");
	printwalllist(n->wl);
	
	if (n->front) {
		lev++;
		for (i=0; i<lev; i++) putchar(' ');
		for (i=0; i<lev; i++) putchar(' ');
		printf("front: ");
		preorderprint(n->front);
		lev--;
	}
	if (n->back) {
		lev++;
		for (i=0; i<lev; i++) putchar(' ');
		for (i=0; i<lev; i++) putchar(' ');
		printf("back: ");
		preorderprint(n->back);
		lev--;
	}
}

void printwalllist(wlelem *wl) {
	wlelem *we;
	int n;
	
	n = 0;
	we = wl->next;

	while(we) {
		printwall(&we->w);
		n++;
		we = we->next;
	}
	printf("%d walls\n", n);
}

void printwall(wall *w) {
	printf("%lf %lf -> %lf %lf\n", w->start[0], w->start[1],
			w->end[0], w->end[1]);
	printf("normal: %lf %lf\n", w->n[0], w->n[1]);
}

void bresenham(int x1, int y1, int x2, int y2, byte c) {
	int dx, dy, x, y, xinc, error;
	int i;
	
	if (y2 < y1) {
		y = y2;
		y2 = y1;
		y1 = y;
		x = x2;
		x2 = x1;
		x1 = x;
	}

	if (x2 > x1) xinc = 1;
	else if (x2 < x1) xinc = -1;
	else {
		for (y = y1; y <= y2; y++) {
			plot(x1, y, c);
		}
		return;
	}
	
	dx = x2 - x1;
	dy = y2 - y1;
	if (dx < 0) dx = -dx;
	error = -dx / 2;
	x = x1;
	y = y1;

	if (!dy) {
		for (i=0; i<=dx; i++) {
			plot(x, y, c);
			x += xinc;
		}
		return;
	}

	for (i=0; i<=dx; i++) {
		plot(x, y, c);
		x += xinc;
		error += dy;
		while (error > 0 && y < y2) {
			y++;
			plot(x, y, c);
			error -= dx;
		}
	}

}

void plot(int x, int y, byte c) {
	if (x < 0 || x >= sc.swidth || y < 0 || y >= sc.sheight) return;
	img[y * sc.swidth + x] = c;
}

void loadtex(int n, char *filename) {
	int i;

	if (!(infile = fopen(filename, "r"))) {
			perror("fopen");
			exit(1);
	}

	readin("%d", &tex[n].width, "texture width");
	readin("%d", &tex[n].height, "texture height");
	tex[n].wmask = tex[n].width - 1;
	tex[n].hmask = tex[n].height - 1;
	for (i=0; i<8; i++) {
		if (tex[n].height & (1 << i)) {
			tex[n].logheight = i;
			break;
		}
	}

	tex[n].data = (byte *) xmalloc(tex[n].width * tex[n].height);
	for (i=0; i<tex[n].width * tex[n].height; i++) {
		readin("%d", &tex[n].data[i], "texture pixel #%d", i);
	}

	fclose(infile);
}

void showtexture(int x, int y, texture_map *t) {
	int i;
	int j;

	for (i=0; i<t->height; i++) {
		for (j=0; j<t->width; j++) {
			img[(y + i) * sc.swidth + x + j] = t->data[i * t->height + j];
		}
	}

	drawimage(img);
}

void orig2wall(origwall *ow, wall *w) {
	vector2 p;

	w->start[0] = (double) ow->start[0];
	w->start[1] = (double) ow->start[1];
	w->end[0] = (double) ow->end[0];
	w->end[1] = (double) ow->end[1];

	w->n[0] = w->end[1] - w->start[1];
	w->n[1] = w->start[0] - w->end[0];
	normalize2(w->n);

	VecVecVec2(p, =, w->end, -, w->start);
	w->length = sqrt(VecDot2(p, p));

	w->plus = ow->plus;
	w->minus = ow->minus;

}

void screenshot(void) {
	static int filenum = 0;
	char filename[100];
	FILE *fp;
	int i, j;

	sprintf(filename, "bsp%d.ben", filenum++);
	fp = fopen(filename, "w");

	if (!fp) {
		perror(filename);
		exit(5);
	}

	for (j=0; j<SHEIGHT; j++) {
		for (i=0; i<SWIDTH; i++) {
			fprintf(fp, "%d\n",img[j*SWIDTH+i]);
		}
	}
	
	fclose(fp);
	sleep(1);
}
