#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
//#include <grx20.h>
#include <assert.h>
#include <graphapp.h>


#define VecLen2(V) sqrt(VecDot2(V, V))
#define VecList2(V) V[0], V[1]
#define VecDot2(A, B) ((A[0]*B[0]) + (A[1]*B[1]))

#define VecScalar2(V, assign_op, k) \
(	V[0] assign_op k, \
	V[1] assign_op k \
)
#define VecVec2(A, assign_op, B) \
(	A[0] assign_op B[0], \
	A[1] assign_op B[1] \
)
#define VecVecVec2(A, assign_op, B, op, C) \
(	A[0] assign_op B[0] op C[0], \
	A[1] assign_op B[1] op C[1] \
)

#define epsilon 0.00001
/*
#define appendwall(a, b) \
	a->next = (wlelem *) salloc(sizeof(wlelem)); \
	a = a->next; \
	a->next = NULL; \
	a->w = b;
	*/

typedef double vector[3];
typedef double vector2[2];

struct wallattr {
	//double red, green, blue;
	double c[3];
	short visible;
};

typedef struct wallattr wallattr;

struct wall {
	vector2 start;
	vector2 end;
	vector2 n;
	wallattr plus, minus;
};

typedef struct wall wall;

struct wlelem {
	struct wlelem *next;
	wall w;
};

typedef struct wlelem wlelem;

struct plane {
	vector2 p;
	vector2 n;
};

typedef struct plane plane;

struct bsptreenode {
	struct bsptreenode *front;
	struct bsptreenode *back;
	plane cut;
	wlelem *wl;
};

typedef struct bsptreenode bsptreenode;

enum {
	drawit = 0
};

struct sceneinfo {
	vector eye;
	vector at;
	vector up;
	double focal;
	double wwidth;
	double wheight;
	int swidth;
	int sheight;
};

typedef struct sceneinfo sceneinfo;

// prototypes
void processkey(drawing d, int key);
void inorderdraw(bsptreenode *r);
void drawscene(void);
void imgplot(int x, int y, double colour[3]);
void bresenham(int x1, int y1, int x2, int y2, double c[3]);
void drawimg(drawing d, rect r);
void getouttahere(window, int, point);
void printwall(wall *);
void preorderprint(bsptreenode *);
void printwalllist(wlelem *);
void makebsptree(bsptreenode *, wlelem *);
void readin(char *, void *, char *, ...);
void parsecomment(void);
void loadfile(char *);

void *salloc(size_t);
void *normalize2(double [2]);
void readinwallattr(wallattr *);

int wallcount;
wlelem *wlhead;
bsptreenode *bsproot;

FILE *infile;

int sheight = 600;
sceneinfo sc;

image img;

int main(int argc, char *argv[]) {
	int i;
	int x1, y1, x2, y2;
	wlelem *we;
	window w;
	drawing d;

	if (argc < 2) {
		fprintf(stderr, "No filename given\n");
		exit(2);
	}

	loadfile(argv[1]);

	bsproot = (bsptreenode *) salloc(sizeof(bsptreenode));
	makebsptree(bsproot, wlhead);

	printf("BSP tree:\n");
	preorderprint(bsproot);
	
	printf("End of BSP tree\n");
	printf("%d\n", wallcount);
	we = wlhead->next;

	while(we) {
		printf("%lf %lf %lf %lf\n", we->w.start[0], we->w.start[1],
				we->w.end[0], we->w.end[1]);
		printf("%lf %lf\n", we->w.n[0], we->w.n[1]);
		we = we->next;
	}

#if 0
	sc.swidth = 600;
	sc.sheight = 600;
	w = newwindow("3D", rect(50,50,sc.swidth,sc.sheight), StandardWindow);
	img = newimage(sc.swidth, sc.sheight, 32);
	drawscene();

	d = newdrawing(rect(0,0,sc.swidth,sc.sheight), drawimg);
	setdata(w, d);	/* store the pointer with the window */
	setmousedown(d, getouttahere);
	setkeyaction(d, processkey);

	show(w);
	mainloop();

	return(0);

#endif

	if (drawit) {
		GrSetMode(GR_width_height_color_graphics, 800, sheight, 16L << 20);
		sheight = GrScreenY();
		GrClearScreen(GrBlack());
		we = wlhead->next;
		while (we) {
			x1 = (int) we->w.start[0];
			y1 = sheight - (int) we->w.start[1];
			x2 = (int) we->w.end[0];
			y2 = sheight - (int) we->w.end[1];
			GrLine(x1, y1, x2, y2, GrWhite());
			x1 = (x1 + x2) / 2;
			y1 = (y1 + y2) / 2;
			if (we->w.plus.visible)
				GrLine(x1, y1, x1 + we->w.n[0] * 10, y1 - we->w.n[1] * 10,
					GrAllocColor(we->w.plus.red * 255, we->w.plus.green*255,
						we->w.plus.blue * 255));
			if (we->w.minus.visible)
				GrLine(x1, y1, x1 - we->w.n[0] * 10, y1 + we->w.n[1] * 10,
					GrAllocColor(we->w.minus.red * 255, we->w.minus.green*255,
						we->w.minus.blue * 255));
			we = we->next;
		}
		getchar();
}

void processkey(drawing d, int key) {
	printf("Received key %d %d\n", key, LEFT);
	switch(key) {
		case LEFT:
			sc.at[0]--;
			drawscene();
			drawimg(d, rect(0,0,sc.swidth,sc.sheight));
			break;

		default:
			break;

	}
}


void drawscene(void) {
	double c[3];
	
	c[0] = c[1] = c[2] = 1;
	imgplot(VecList2(sc.at), c);
	inorderdraw(bsproot);
}

void inorderdraw(bsptreenode *r) {
	wlelem *we;
	double c[3];
	int x1, x2, y1, y2;
	
	if (r->front) {
		inorderdraw(r->front);
	}
	if (r->back) {
		inorderdraw(r->back);
	}

	c[0] = 1;
	c[1] = 1;
	c[2] = 1;

	we = r->wl->next;
	while (we) {
		x1 = (int) we->w.start[0];
		y1 = sheight - (int) we->w.start[1];
		x2 = (int) we->w.end[0];
		y2 = sheight - (int) we->w.end[1];
		bresenham(x1, y1, x2, y2, c);
		x1 = (x1 + x2) / 2;
		y1 = (y1 + y2) / 2;
		if (we->w.plus.visible)
			bresenham(x1, y1, x1 + (int) we->w.n[0] * 10,
					y1 - (int) we->w.n[1] * 10, we->w.plus.c);
		if (we->w.minus.visible)
			bresenham(x1, y1, x1 - (int) we->w.n[0] * 10,
					y1 + (int) we->w.n[1] * 10, we->w.minus.c);
		we = we->next;
	}
}

void drawimg(drawing d, rect r) {
	drawimage(img, r, getrect(img));
}

void getouttahere(window w, int buttons, point xy) {
        if (buttons) {
                exitapp();
        } else {
                printf("Bug? Mouse down called for no reason\n");
        }
}

void loadfile(char *filename) {
	int i, j, k;
	wlelem *we;
	wall *wp;

	if (!(infile = fopen(filename, "r"))) {
			perror("fopen");
			exit(1);
	}

	readin("%d", &wallcount, "number of walls");

	wlhead = (wlelem *) salloc(sizeof(wlelem));
	wlhead->next = NULL;

	we = wlhead;
	for (i=0; i<wallcount; i++) {
		we->next = (wlelem *) salloc(sizeof(wlelem));
		we = we->next;
		we->next = NULL;
		wp = &we->w;
		readin("%lf", &wp->start[0], "wall #%d start x", i);
		readin("%lf", &wp->start[1], "wall #%d start y", i);
		readin("%lf", &wp->end[0], "wall #%d end x", i);
		readin("%lf", &wp->end[1], "wall #%d end y", i);
		readin("%lf", &wp->n[0], "wall #%d normal x", i);
		readin("%lf", &wp->n[1], "wall #%d normal y", i);
		readinwallattr(&wp->plus);
		readinwallattr(&wp->minus);
	}

	readin("%lf", &sc.at[0], "player start x");
	readin("%lf", &sc.at[1], "player start y");
	sc.at[2] = 0;
}

void readinwallattr(wallattr *wattr) {
	int i;

	readin("%hd", &wattr->visible, "visibility flag");

	if (wattr->visible) {
		readin("%lf", &wattr->c[0], "wall #%d side red", i);
		readin("%lf", &wattr->c[1], "wall #%d side green", i);
		readin("%lf", &wattr->c[2], "wall #%d side blue", i);
	}
}

void readin(char *s, void *f, char *format, ...) {
	va_list arg;
	char done;
	int check;
	char ch;

	done = 0;
	do {
		if (feof(infile)) {
			va_start(arg, format);
			fprintf(stderr, "Unexpected end of file.\nError reading ");
			vfprintf(stderr, format, arg);
			fprintf(stderr, "\n");
			va_end(arg);
			exit(1);
		}

		fscanf(infile, " %c", &ch);
		if (ch == '/') {
			parsecomment();
		} else {
			ungetc(ch, infile);
			check = fscanf(infile, s, f);
			if (!check) {
				va_start(arg, format);
				fprintf(stderr, "Error reading ");
				vfprintf(stderr, format, arg);
				fprintf(stderr, "\n");
				va_end(arg);
				exit(1);
			}
			//printf("Read %s\n", thing);
			done++;
		}
	} while (!done);
}

void parsecomment(void) {
	char ch;
	char flag;

	if ((ch = getc(infile) == '/')) {
		while ((ch = getc(infile) != '\n'));
		return;
	} else if ((ch = getc(infile)) != '*') {
		fprintf(stderr, "Warning: bad comment\n");
	}
	flag = 0;
	do {
		do {
			ch = getc(infile);
		} while (ch != '*');
		if ((ch = getc(infile)) == '/') {
			flag++;
		}
	} while (!flag);
}

void *salloc(size_t size) {
	void *ptr;
	
	ptr = (void *) malloc(size);
	if (!ptr) {
		fprintf(stderr, "out of memory\n");
		exit(2);
	}

	return(ptr);
}

void *normalize2(double v[2]) {
	double n;

	n = VecLen2(v); //sqrt(v[0] * v[0] + v[1] * v[1]);
	VecScalar2(v, /=, n);
	/*
	v[0] /= n;
	v[1] /= n;
	*/
}

void makebsptree(bsptreenode *node, wlelem *wl) {
	wlelem *wlfronthead, *wlbackhead;
	wlelem *wefront, *weback, *weon, *we;
	wlelem *newwe;
	double d, d2, mu;
	vector2 p, dir, v;
	char type;

	wefront = wlfronthead = (wlelem *) salloc(sizeof(wlelem));
	weback = wlbackhead = (wlelem *) salloc(sizeof(wlelem));
	wefront->next = weback->next = NULL;
	node->wl = (wlelem *) salloc(sizeof(wlelem));
	node->wl->next = NULL;

	weon = node->wl;

	we = wl->next;

	assert(we);
	if (!we) {
		return;
	}
	printf("Cut: ");
	printwall(&we->w);

	VecVec2(node->cut.p, =, we->w.start);
	VecVec2(node->cut.n, =, we->w.n);
	/*
	weon->next = (wlelem *) salloc(sizeof(wlelem));
	weon = weon->next;
	weon->next = NULL;
	weon->w = we->w;
	*/
	weon->next = we;
	weon = we;
	we = we->next;
	weon->next = NULL;

	while(we) {
		VecVecVec2(p, =, we->w.start, -, node->cut.p);
		d = VecDot2(node->cut.n, p);
		VecVecVec2(p, =, we->w.end, -, node->cut.p);
		d2 = VecDot2(node->cut.n, p);
		// 1 = in front, -1 = behind, 0 = on, 2 = cross
		if (d > epsilon) {
			if (d2 >= -epsilon) type = 1;
			else type = 2;
		} else if (fabs(d) < epsilon) {
			if (d2 > epsilon) type = 1;
			else if (d2 < -epsilon) type = -1;
			else type = 0;
		} else {
			if (d2 <= epsilon) type = -1;
			else type = 2;
		}
		switch(type) {
			case 0:
				weon->next = we;
				weon = we;
				we = we->next;
				weon->next = NULL;
				/*
				weon->next = (wlelem *) salloc(sizeof(wlelem));
				weon = weon->next;
				weon->next = NULL;
				weon->w = we->w;
				*/
				break;

			case 1:
				wefront->next = we;
				wefront = we;
				we = we->next;
				wefront->next = NULL;
				break;
				
			case -1:
				weback->next = we;
				weback = we;
				we = we->next;
				weback->next = NULL;
				break;

			case 2:
				// cut the wall
				// intersection of 2 lines
				VecVecVec2(dir, =, we->w.end, -, we->w.start);
				VecVecVec2(v, =, node->cut.p, -, we->w.start);
				mu = VecDot2(node->cut.n, v) / VecDot2(node->cut.n, dir);
				VecVecVec2(p, =, we->w.start, +, mu*dir);
				VecVecVec2(v, =, node->cut.p, -, p);
				printf("Intersection %lf, %lf\n", VecList2(p));
				printf("should be 0: %d\n", VecDot2(node->cut.n, v));
				
				newwe = (wlelem *) salloc(sizeof(wlelem));
				newwe->w = we->w;
				VecVec2(newwe->w.start, =, p);
				VecVec2(we->w.end, =, p);
				if (d > epsilon) {
					wefront->next = we;
					wefront = we;
					we = we->next;
					wefront->next = NULL;
					weback->next = newwe;
					weback = newwe;
					weback->next = NULL;
				} else {
					weback->next = we;
					weback = we;
					we = we->next;
					weback->next = NULL;
					wefront->next = newwe;
					wefront = newwe;
					wefront->next = NULL;
				}
				break;

			default:
				we = we->next;
				break;
		}
	
	}
	if (wlfronthead->next) {
		node->front = (bsptreenode *) salloc(sizeof(bsptreenode));
		makebsptree(node->front, wlfronthead);
	} else {
		node->front = NULL;
	}
	if (wlbackhead->next) {
		node->back = (bsptreenode *) salloc(sizeof(bsptreenode));
		makebsptree(node->back, wlbackhead);
	} else {
		node->back = NULL;
	}
}

void preorderprint(bsptreenode *n) {
	static int lev = 0;
	int i;

	assert(n);
	
	printf("this: ");
	printwalllist(n->wl);
	
	if (n->front) {
		lev++;
		for (i=0; i<lev; i++) putchar(' ');
		for (i=0; i<lev; i++) putchar(' ');
		printf("front: ");
		preorderprint(n->front);
		lev--;
	}
	if (n->back) {
		lev++;
		for (i=0; i<lev; i++) putchar(' ');
		for (i=0; i<lev; i++) putchar(' ');
		printf("back: ");
		preorderprint(n->back);
		lev--;
	}
}

void printwalllist(wlelem *wl) {
	wlelem *we;
	int n;
	
	n = 0;
	we = wl->next;

	while(we) {
		printwall(&we->w);
		n++;
		we = we->next;
	}
	printf("%d walls\n", n);
}

void printwall(wall *w) {
	printf("%lf %lf -> %lf %lf\n", w->start[0], w->start[1],
			w->end[0], w->end[1]);
	printf("normal: %lf %lf\n", w->n[0], w->n[1]);
}

void bresenham(int x1, int y1, int x2, int y2, double c[3]) {
	int dx, dy, x, y, xinc, error;
	int i;
	
	if (y2 < y1) {
		y = y2;
		y2 = y1;
		y1 = y;
		x = x2;
		x2 = x1;
		x1 = x;
	}

	if (x2 > x1) xinc = 1;
	else if (x2 < x1) xinc = -1;
	else {
		for (y = y1; y <= y2; y++) {
			imgplot(x1, y, c);
		}
		return;
	}
	
	dx = x2 - x1;
	dy = y2 - y1;
	if (dx < 0) dx = -dx;
	error = -dx / 2;
	x = x1;
	y = y1;

	if (!dy) {
		for (i=0; i<=dx; i++) {
			imgplot(x, y, c);
			x += xinc;
		}
		return;
	}

	for (i=0; i<=dx; i++) {
		imgplot(x, y, c);
		x += xinc;
		error += dy;
		while (error > 0 && y < y2) {
			y++;
			imgplot(x, y, c);
			error -= dx;
		}
	}

}

void imgplot(int x, int y, double colour[3]) {
	if (y >= 0 && y < sc.sheight && x >= 0 && x < sc.swidth) {
		y *= sc.swidth*4;
		y += x*4;
		img->pixels[y] = 0;
		img->pixels[y + 1] = (byte) (colour[0]*255);
		img->pixels[y + 2] = (byte) (colour[1]*255);
		img->pixels[y + 3] = (byte) (colour[2]*255);
	}
}
