#include <stdio.h>
#include <stdlib.h>
#include "graphapp.h"

struct sceneinfo {
	double eye[3];
	double at[3];
	double up[3];
	double focal;
	double wwidth;
	double wheight;
	int swidth;
	int sheight;
};

struct objstruct {
	int vcount;			// no. of vertices
	double (*v)[3];		// vertices
	int fcount;			// no. of faces
	int *ecount;		// no. of edges on a face
	int (**e)[2];		// edges of a face
	double (*n)[3];		// normal
	double *r;
	double *g;
	double *b;
};

struct ilist {
	struct ilist *next;
	int i;
};

void readin(char *, void *);
void parsecomment(void);
void scanfill(struct objstruct *o, int f);

double amblight[3];
int lightcount;
int objcount;
int vtotal;

FILE *infile;

int main(int argc, char **argv) {
	int i, j, k;
	double f;
	byte r, g, b;
	struct sceneinfo s;
	struct objstruct *o;

	if (!(infile = fopen("tt", "r"))) { //stage1.1.gen", "r"))) {
			perror("fopen");
			exit(1);
	}

	readin("%lf", &s.eye[0]);
	readin("%lf", &s.eye[1]);
	readin("%lf", &s.eye[2]);
	readin("%lf", &s.at[0]);
	readin("%lf", &s.at[1]);
	readin("%lf", &s.at[2]);
	readin("%lf", &s.up[0]);
	readin("%lf", &s.up[1]);
	readin("%lf", &s.up[2]);
	readin("%lf", &s.focal);
	readin("%lf", &s.wwidth);
	readin("%lf", &s.wheight);
	readin("%d", &s.swidth);
	readin("%d", &s.sheight);

	readin("%lf", &amblight[0]);
	readin("%lf", &amblight[1]);
	readin("%lf", &amblight[2]);

	/* TODO: Actually read where the lights are */
	readin("%d", &lightcount);
	for (i=0; i<lightcount; i++) {
		for (j=0; j<6; j++) {
			readin("%lf", &f);
		}
	}

	readin("%d", &objcount);
	o = (struct objstruct *) malloc(sizeof(struct objstruct) * objcount);

	readin("%d", &vtotal);

	for (i=0; i<objcount; i++) {

		readin("%d", &o[i].vcount);
		o[i].v = (double (*)[3]) malloc(sizeof(double) * 3 * o[i].vcount);
		for (j=0; j<o[i].vcount; j++) {
			readin("%lf", &o[i].v[j][0]);
			readin("%lf", &o[i].v[j][1]);
			readin("%lf", &o[i].v[j][2]);
		}

		readin("%d", &o[i].fcount);
		o[i].ecount = (int *) malloc(sizeof(int) * o[i].fcount);
		o[i].n = (double (*)[3]) malloc(sizeof(double) * 3 * o[i].fcount);
		o[i].e = (int (**)[2]) malloc(sizeof(int (**)[2]) * o[i].fcount);
		o[i].r = (double *) malloc(sizeof(double) * o[i].fcount);
		o[i].g = (double *) malloc(sizeof(double) * o[i].fcount);
		o[i].b = (double *) malloc(sizeof(double) * o[i].fcount);
		
		for (j=0; j<o[i].fcount; j++) {
			/* Memory inefficient - may change to linked list eventually
			 * (no. of edges <= no. of vertices)
			 */
			o[i].e[j] = (int (*)[2]) malloc(sizeof(int) * 2 * o[i].vcount);


			k = 0;
			do {
				readin("%d", &o[i].e[j][k][0]);

				if (o[i].e[j][k][0] != -1) {
					readin("%d", &o[i].e[j][k][1]);
					k++;
				}
			} while (o[i].e[j][k][0] != -1);
			o[i].ecount[j] = k;

			readin("%lf", &o[i].n[j][0]);
			readin("%lf", &o[i].n[j][1]);
			readin("%lf", &o[i].n[j][2]);

			readin("%lf", &o[i].r[j]);
			readin("%lf", &o[i].g[j]);
			readin("%lf", &o[i].b[j]);

			readin("%lf", &f);
			readin("%lf", &f);
			readin("%lf", &f);
		}

	}

	init();
	getchar();
	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			scanfill(&o[i], j);
		}
	}

	getchar();
	return(0);
}

void scanfill(struct objstruct *o, int f) {
	int i, j;
	int vmin;
	int ymin, ymax;
	int max;
	int (*sv)[2];	// Screen coords of vertices
	char *dir;
	struct ilist *inter, *ip, *temp;
	int error;
	int endx, endy;
	int x, y, dx, dy, x2, y2;

	sv = (int (*)[2]) malloc(sizeof(int) * 2 * o->ecount[f]);
	dir = (char *) malloc(sizeof(char) * o->vcount);

	for (i=0; i<o->vcount; i++) {
		dir[i] = 0;
		sv[i][0] = o->v[i][0];
		sv[i][1] = o->v[i][1];
	}

	/* Find minimum numbered vertex */
	vmin = o->e[f][0][0];
	for (i=0; i<o->ecount[f]; i++) {
		if (vmin > o->e[f][i][0]) vmin = o->e[f][i][0];
		if (vmin > o->e[f][i][1]) vmin = o->e[f][i][1];
	}

	/* Renumber consistently */
	for (i=0; i<o->ecount[f]; i++) {
		if (o->e[f][i][0] > o->e[f][i][1]) {
			if (o->e[f][i][1] != vmin) {
				j = o->e[f][i][1];
				o->e[f][i][1] = o->e[f][i][0];
				o->e[f][i][0] = j;
			}
		}
	}

	/* Find min, max y coords */
	ymax = ymin = sv[o->e[f][0][0]][1];
	for (i=0; i<o->ecount[f]; i++) {
		if (ymax < sv[o->e[f][i][0]][1]) {
			ymax = sv[o->e[f][i][0]][1];
		}
		if (ymin > sv[o->e[f][i][0]][1]) {
			ymin = sv[o->e[f][i][0]][1];
		}
	}

	max = ymax - ymin + 1;
	inter = (struct ilist *) malloc(sizeof(struct ilist) * max);
	for (i=0; i<max; i++) {
		inter[i].next = NULL;
	}

	/* Calculate intersections */

	/*
	for (i=0; i<o[0].ecount[0]; i++) {
		printf("%d - %d\n", o[0].e[0][i][0], o[0].e[0][i][1]);
	}
	printf("%d %d %d\n", vmin, ymin, ymax);
	*/

	/*
	for (i=0; i<o->ecount[f]; i++) {
		drawline(pt(sv[o->e[f][i][0]][0], sv[o->e[f][i][0]][1]),
				pt(sv[o->e[f][i][1]][0], sv[o->e[f][i][1]][1]));
	}
	*/
	for (i=0; i<o->ecount[f]; i++) {
		x = sv[o->e[f][i][0]][0];
		y = sv[o->e[f][i][0]][1];
		endx = sv[o->e[f][i][1]][0];
		endy = sv[o->e[f][i][1]][1];
		drawline(pt(x, y), pt(endx, endy));
		dx = endx - x;
		dy = endy - y;
		if (dx > 0) x2 = 1;
		else {
			dx = -dx;
			x2 = -1;
		}
		if (dy > 0) y2 = 1;
		else {
			dy = -dy;
			y2 = -1;
		}

		if (dy) {
			dir[o->e[f][i][0]] += y2;
			dir[o->e[f][i][1]] -= y2;

			error = dy;
			if (error > 0) error = -error;
			error /= 2;
			do {

				y-=ymin;
				if (!inter[y].next) {
					inter[y].next = (struct ilist *) malloc(sizeof(struct ilist));
					inter[y].next->i = x;
					inter[y].next->next = NULL;
				} else {
					ip = &inter[y];
					while (ip->next && ip->next->i < x)
						ip = ip->next;

					temp = ip->next;
					ip->next = (struct ilist *) malloc(sizeof(struct ilist));
					ip->next->i = x;
					ip->next->next = temp;
				}
				y+=ymin;

				y += y2;
				error += dx;
				while (error > 0) {
					x += x2;
					error -= dy;
				}
			} while(y != endy);

			/* More efficient if I duplicate code ? */
				y-=ymin;
				if (!inter[y].next) {
					inter[y].next = (struct ilist *) malloc(sizeof(struct ilist));
					inter[y].next->i = x;
					inter[y].next->next = NULL;
				} else {
					ip = &inter[y];
					while (ip->next && ip->next->i < x)
						ip = ip->next;

					temp = ip->next;
					ip->next = (struct ilist *) malloc(sizeof(struct ilist));
					ip->next->i = x;
					ip->next->next = temp;
				}

		}
	}

	for (i=0; i<o->ecount[f]; i++) {
		vmin = o->e[f][i][0];
		if (!dir[vmin]) {
			y = sv[vmin][1] - ymin;
			ip = &inter[y];
			while (ip->next->i != sv[vmin][0]) ip = ip->next;
			temp = ip->next->next;
			free(ip->next);
			ip->next = temp;

		}
	}

	/* Draw the scan lines */
	for(i=0; i<max; i++) {
		ip = &inter[i];
		while (ip->next) {
			ip = ip->next;
			x = ip->i;
			if (!ip->next) continue;
			ip = ip->next;
			x2 = ip->i;
			//printf("%d %d\n", x, x2);
			//GrLine(x, i + ymin, x2, i + ymin);
			drawline(pt(x, i + ymin), pt(x2, i + ymin));
		}
	}

	free(sv);
}

void readin(char *s, void *f) {
	char done;
	int check;
	char ch;

	done = 0;
	do {
		fscanf(infile, " %c", &ch);
		if (ch == '/') {
			parsecomment();
		} else {
			ungetc(ch, infile);
			check = fscanf(infile, s, f);
			if (!check) {
				fprintf(stderr, "Bad number\n");
				exit(1);
			}
			done++;
		}
	} while (!done);
}

void parsecomment(void) {
	char ch;
	char flag;

	if ((ch = getc(infile)) != '*') {
		fprintf(stderr, "Warning: bad comment\n");
	}
	flag = 0;
	do {
		do {
			ch = getc(infile);
		} while (ch != '*');
		if ((ch = getc(infile)) == '/') {
			flag++;
		}
	} while (!flag);
}
