/* Ray tracer
 * 25/8/98
   Ben Lin
   
   It might be easier if I preprocess all the info so the faces always
   look like 0 1 1 2 2 3 3 0 or something - plus I get to check for errors
   */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "graphapp.h"

#define normvec(a) dotvec(a, a)
#define lenvec(a) sqrt(dotvec(a, a))

struct sceneinfo {
	double eye[3];
	double at[3];
	double up[3];
	double focal;
	double wwidth;
	double wheight;
	int swidth;
	int sheight;
};

struct facestruct {
	//struct objstruct *o;	// owner
	int ecount;
	int (*e)[2];
	double n[3];	//	normal
	double sn[3];	// normal after transformation
	double colour[3];
	double diffuse, spec;
	int specpower;

	//	extra info:
	int vmin;	// smallest-numbered vertex
	char hidden;	//self-hidden
};

struct objstruct {
	int vcount;			// no. of vertices
	//double (*v)[3];		// vertices
	int fcount;			// no. of faces
	struct facestruct *face;
};

struct ilist {
	struct ilist *next;
	int i;
};

struct lightstruct {
	double loc[3];	// location
	double colour[3];
};

void readin(char *, void *, char *);
void parsecomment(void);
void scanfill(struct objstruct *, struct facestruct *);
//void draw_scene(drawing, rect);
void draw_scene(void);
void loadfile(char *);
void transscene(void);
void render(void);
void raytrace(double *, double *);
double findinter(struct facestruct *, double *);
void outline(struct facestruct *);

void showvec(double *);
void addvec(double *, double *, double *);
void subvec(double *, double *, double *);
void multvec(double, double *);
void matvec(double *, double [3][3], double *);
void matmat(double [3][3], double [3][3], double [3][3]);
void showmat(double [3][3]);
double dotvec(double *, double *);

double amblight[3];
int lightcount;
int objcount;
int vcount;
struct sceneinfo s;
struct objstruct *o;
struct lightstruct *light;
double (*vertex)[3];
double (*svertex)[3];	// vertex coords after translation and rotation
double seye;		// eye z coord after transformation; at and up are trivial

FILE *infile;

int main(int argc, char **argv) {

#if 0
	window w;
	drawing d;
#endif

	loadfile("tt");
	transscene();

#if 0
	w = newwindow("Stage1", rect(50,50,s.swidth,s.sheight), StandardWindow);
	d = newdrawing(rect(0,0,s.swidth,s.sheight), draw_scene);
	setdata(w, d);	/* store the pointer with the window */

	show(w);
	mainloop();
#endif
	getchar();
	init();
	render();
	getchar();
	/*
	getchar();
	draw_scene();
	getchar();
	*/

	return(0);
}

/* Ray trace */

void render(void) {
	int i, j;
	double ray[3], v[3];
	double colour[3];

	ray[0] = ray[1] = 0;
	ray[2] = seye;

	// Eliminate self-hidden faces
	// Doesn't seem to work
	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			subvec(v, ray, svertex[o[i].face[j].e[0][0]]);
			if (dotvec(v, o[i].face[j].sn) < 0) {
				o[i].face[j].hidden = 1;
				//printf("hidden\n");
			} else {
				o[i].face[j].hidden = 0;
			outline(&o[i].face[j]);
				//printf("visible\n");

			}

		}
	}
	
	ray[0] = s.swidth / 2;
	ray[1] = s.sheight / 2;
	ray[2] = -seye;

	for (j=0; j<s.sheight; j++, ray[1]++) {
		for (i=0; i<s.swidth; i++, ray[0]++) {
			// For now, it'll be really inefficient
			// i.e. a big loop checking each face on each object
			// (later on I'll have bounding spheres or something)
			raytrace(colour, ray);
			//setcolour(initrgb((int) (colour[0]*255), (int) (colour[1]*255), (int) (colour[2]*255)));
			//drawpoint(pt(i, j));
		}
	}
}

void raytrace(double *colour, double *ray) {
	int i, j;
	double mu;

	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			mu = findinter(&o[i].face[j], ray);
		}
	}
}

double findinter(struct facestruct *f, double *ray) {
	double s;

	s = dotvec(f->n, svertex[f->e[0][0]]);
	return s;
}

/* Parse a .gen file */

void loadfile(char *filename) {
	int i, j, k;
	int v;
	int n;
	double f;

	if (!(infile = fopen(filename, "r"))) {
			perror("fopen");
			exit(1);
	}

	readin("%lf", &s.eye[0], "eye x");
	readin("%lf", &s.eye[1], "eye y");
	readin("%lf", &s.eye[2], "eye z");
	readin("%lf", &s.at[0], "view at x");
	readin("%lf", &s.at[1], "view at y");
	readin("%lf", &s.at[2], "view at z");
	readin("%lf", &s.up[0], "view up x");
	readin("%lf", &s.up[1], "view up y");
	readin("%lf", &s.up[2], "view up z");
	readin("%lf", &s.focal, "focal length");
	readin("%lf", &s.wwidth, "window width");
	readin("%lf", &s.wheight, "window height");
	readin("%d", &s.swidth, "screen height");
	readin("%d", &s.sheight, "screen height");

	readin("%lf", &amblight[0], "ambient light r");
	readin("%lf", &amblight[1], "ambient light g");
	readin("%lf", &amblight[2], "ambient light b");

	readin("%d", &lightcount, "number of lights");

	light = (struct lightstruct *) malloc(sizeof(struct lightstruct) * lightcount);

	for (i=0; i<lightcount; i++) {
		readin("%lf", &light[i].loc[0], "light x");
		readin("%lf", &light[i].loc[1], "light y");
		readin("%lf", &light[i].loc[2], "light z");

		readin("%lf", &light[i].colour[0], "light r");
		readin("%lf", &light[i].colour[1], "light g");
		readin("%lf", &light[i].colour[2], "light b");
	}

	readin("%d", &objcount, "number of objects");
	o = (struct objstruct *) malloc(sizeof(struct objstruct) * objcount);

	readin("%d", &vcount, "total vertices");
	vertex = (double (*)[3]) malloc(sizeof(double) * 3 * vcount);
	svertex = (double (*)[3]) malloc(sizeof(double) * 3 * vcount);
	v = 0;

	for (i=0; i<objcount; i++) {

		readin("%d", &o[i].vcount, "number of vertices in object");

		//o[i].v = (double (*)[3]) malloc(sizeof(double) * 3 * o[i].vcount);
		for (j=0; j<o[i].vcount; j++) {
			readin("%lf", &vertex[v][0], "vertex x");
			readin("%lf", &vertex[v][1], "vertex y");
			readin("%lf", &vertex[v][2], "vertex z");
			v++;
		}

		readin("%d", &o[i].fcount, "number of faces");
		o[i].face = (struct facestruct *) malloc(sizeof(struct facestruct) * o[i].fcount);
		
		printf("object %d\n", i);
		for (j=0; j<o[i].fcount; j++) {
			printf("\tface %d\n", j);
			/* Memory inefficient - may change to linked list eventually
			 * (no. of edges <= no. of vertices)
			 */
			o[i].face[j].e = (int (*)[2]) malloc(sizeof(int) * 2 * o[i].vcount);

			o[i].face[j].vmin = o[i].vcount;
			k = 0;
			do {
				readin("%d", &n, "edge's from-vertex");

				if (n != -1) {
					if (o[i].face[j].vmin > n) o[i].face[j].vmin = n;
					o[i].face[j].e[k][0] = v - o[i].vcount + n;
					readin("%d", &n, "edge's to-vertex");
					if (o[i].face[j].vmin > n) o[i].face[j].vmin = n;
					o[i].face[j].e[k][1] = v - o[i].vcount + n;
					k++;
				}
			} while (n != -1);
			o[i].face[j].ecount = k;

			/* Renumber consistently */
			for (k=0; k<o[i].face[j].ecount; k++) {
				if ((o[i].face[j].e[k][0] > o[i].face[j].e[k][1])
						&& (o[i].face[j].e[k][1] != o[i].face[j].vmin)) {
					n = o[i].face[j].e[k][1];
					o[i].face[j].e[k][1] = o[i].face[j].e[k][0];
					o[i].face[j].e[k][0] = n;
				}
			}

			readin("%lf", &o[i].face[j].n[0], "normal x");
			readin("%lf", &o[i].face[j].n[1], "normal y");
			readin("%lf", &o[i].face[j].n[2], "normal z");

			/* Colour */
			readin("%lf", &o[i].face[j].colour[0], "face r");
			readin("%lf", &o[i].face[j].colour[1], "face g");
			readin("%lf", &o[i].face[j].colour[2], "face b");

			/* Diffuse, spec(ular), spec. power */
			readin("%lf", &o[i].face[j].diffuse, "diffuse");
			readin("%lf", &o[i].face[j].spec, "specular");
			readin("%d", &o[i].face[j].specpower, "specular power");

		}

	}

}

//void draw_scene(drawing d, rect r) {
void draw_scene(void) {
	int i, j;

	setcolour(Black);
	fillrect(initrect(0, 0, s.swidth, s.sheight));

	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			scanfill(&o[i], &o[i].face[j]);
		}
	}
}

// Don't need objstruct *o, for now
void scanfill(struct objstruct *o, struct facestruct *f) {
	int i, j;
	int ymin, ymax;
	int max;
	int (*sv)[2];	// Screen coords of vertices
	char *dir;	// Records type of vertex - peak or not
	struct ilist *inter, *ip, *temp;
	int error;
	int x, y, dx, dy, x2, y2;
	int endx, endy;

	setcolour(initrgb((int) (f->colour[0]*255), (int) (f->colour[1]*255), (int) (f->colour[2]*255)));

	sv = (int (*)[2]) malloc(sizeof(int) * 2 * vcount);
	dir = (char *) malloc(sizeof(char) * vcount);

	for (i=0; i<vcount; i++) {
		dir[i] = 0;
		sv[i][0] = (int) vertex[i][0];
		sv[i][1] = (int) vertex[i][1];
	}

	/* Find min, max y coords */
	ymax = ymin = sv[f->e[0][0]][1];
	for (i=0; i<f->ecount; i++) {
		if (ymax < sv[f->e[i][0]][1]) {
			ymax = sv[f->e[i][0]][1];
		}
		if (ymin > sv[f->e[i][0]][1]) {
			ymin = sv[f->e[i][0]][1];
		}
	}

	max = ymax - ymin + 1;
	inter = (struct ilist *) malloc(sizeof(struct ilist) * max);
	for (i=0; i<max; i++) {
		inter[i].next = NULL;
	}

	/* Calculate intersections */

/*
	for (i=0; i<o[0].ecount[0]; i++) {
		printf("%d - %d\n", o[0].e[0][i][0], o[0].e[0][i][1]);
	}
	printf("%d %d %d\n", vmin, ymin, ymax);
*/

	/*
	for (i=0; i<f->ecount; i++) {
		drawline(pt(sv[f->e[f][i][0]][0], sv[o->e[i][0]][1]),
				pt(sv[f->e[f][i][1]][0], sv[o->e[i][1]][1]));
	}
	*/
	for (i=0; i<f->ecount; i++) {
		x = sv[f->e[i][0]][0];
		y = sv[f->e[i][0]][1];
		endx = sv[f->e[i][1]][0];
		endy = sv[f->e[i][1]][1];
		drawline(pt(x, y), pt(endx, endy));
		//printf("%d %d, %d %d", x, y, endx, endy);
		dx = endx - x;
		dy = endy - y;
		if (dx > 0) x2 = 1;
		else {
			dx = -dx;
			x2 = -1;
		}
		if (dy > 0) {
			y2 = 1;
		} else {
			dy = -dy;
			y2 = -1;
		}

		if (dy) {
			dir[f->e[i][0]] += y2;
			dir[f->e[i][1]] -= y2;
			error = dy;
			if (error > 0) error = -error;
			error /= 2;
			y-=ymin;
			endy-=ymin;
			do {

				if (!inter[y].next) {
					inter[y].next = (struct ilist *) malloc(sizeof(struct ilist));
					inter[y].next->i = x;
					inter[y].next->next = NULL;
				} else {
					ip = &inter[y];
					while (ip->next && ip->next->i < x)
						ip = ip->next;

					temp = ip->next;
					ip->next = (struct ilist *) malloc(sizeof(struct ilist));
					ip->next->i = x;
					ip->next->next = temp;
				}

				y += y2;

				error += dx;
				while (error > 0) {
					x += x2;
					error -= dy;
				}
			} while(y != endy);

			/* Must decide whether to include this vertex or not */

			/* More efficient if I duplicate code ? */
				if (!inter[y].next) {
					inter[y].next = (struct ilist *) malloc(sizeof(struct ilist));
					inter[y].next->i = x;
					inter[y].next->next = NULL;
				} else {
					ip = &inter[y];
					while (ip->next && ip->next->i < x)
						ip = ip->next;

					temp = ip->next;
					ip->next = (struct ilist *) malloc(sizeof(struct ilist));
					ip->next->i = x;
					ip->next->next = temp;
				}

		}
	}

	for(i=0; i<f->ecount; i++) {
		/* get vertex number of from-vertex of this edge */
		j = f->e[i][0];
		if (!dir[j]) {
			/* Must remove a vertex in this case */
			y = sv[j][1] - ymin;
			ip = &inter[y];
			while (ip->next->i != sv[j][0]) ip = ip->next;
			temp = ip->next->next;
			free(ip->next);
			ip->next = temp;
		}

	}

	/* Draw the scan lines */
	for(i=0; i<max; i++) {
		ip = &inter[i];
		while (ip->next) {
			ip = ip->next;
			x = ip->i;
			if (!ip->next) continue;
			ip = ip->next;
			x2 = ip->i;
			//printf("%d %d\n", x, x2);
			drawline(pt(x, i + ymin), pt(x2, i + ymin));
		}
	}

	free(sv);
}

void outline(struct facestruct *f) {
	int i, j;
	double x1, y1, x2, y2;

	setcolour(Blue);
	for (i=0; i<f->ecount; i++) {
		x1 = svertex[f->e[i][0]][0];
		x1 *= seye / (seye - svertex[f->e[i][0]][2]);
		x1 += s.swidth / 2;
		y1 = svertex[f->e[i][0]][1];
		y1 *= seye / (seye - svertex[f->e[i][0]][2]);
		y1 += s.sheight / 2;
		x2 = svertex[f->e[i][1]][0];
		x2 *= seye / (seye - svertex[f->e[i][1]][2]);
		x2 += s.swidth / 2;
		y2 = svertex[f->e[i][1]][1];
		y2 *= seye / (seye - svertex[f->e[i][1]][2]);
		y2 += s.sheight / 2;
		//printf("%lf %lf, %lf %lf\n", x1, y1, x2, y2);
		drawline(pt((int) x1, (int) y1), pt((int) x2, (int) y2));

	}
}

void readin(char *s, void *f, char *thing) {
	char done;
	int check;
	char ch;

	done = 0;
	do {
		fscanf(infile, " %c", &ch);
		if (ch == '/') {
			parsecomment();
		} else {
			ungetc(ch, infile);
			check = fscanf(infile, s, f);
			if (!check) {
				fprintf(stderr, "Error reading %s\n", thing);
				exit(1);
			}
			//printf("Read %s\n", thing);
			done++;
		}
	} while (!done);
}

void parsecomment(void) {
	char ch;
	char flag;

	if ((ch = getc(infile)) != '*') {
		fprintf(stderr, "Warning: bad comment\n");
	}
	flag = 0;
	do {
		do {
			ch = getc(infile);
		} while (ch != '*');
		if ((ch = getc(infile)) == '/') {
			flag++;
		}
	} while (!flag);
}

// Transform the coords of each vertex so that the camera faces the z=0 plane
void transscene(void) {
	double tr[3];
	double rot[3][3];
	double scale[3][3];
	double combo[3][3];
	double combo2[3][3];
	double v[3];
	double v2[3];
	int i, j;
	double f, g;

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			rot[i][j] = scale[i][j] = 0;
		}
	}

	// scale x and y coords accordingly, and scale z averagely
	scale[0][0] = s.swidth / s.wwidth;
	scale[1][1] = s.sheight / s.wheight;
	scale[2][2] = (scale[0][0] + scale[1][1]) / 2;
	printf("Scaling matrix is:\n");
	showmat(scale);
	printf("\n");

	// want the camera to be facing, and focused on the z=0 plane
	// work out what should go to (0, 0, 0)

	subvec(tr, s.at, s.eye);
	subvec(v, s.at, s.eye);
	/*
	showvec(v);
	printf("%lf\n\n", lenvec(v));
	*/
	multvec(1 - (s.focal / lenvec(v)), tr);
	subvec(tr, tr, s.at);
	printf("Translating by");
	showvec(tr);
	printf("\n");

	// rotate about x-axis so that eye lies in x-z plane
	rot[0][0] = 1;
	f = sqrt(v[2] * v[2] + v[1] * v[1]);
	rot[1][1] = rot[2][2] = -v[2] / f;
	rot[2][1] = -(rot[1][2] = v[1] / f);

	matmat(combo, scale, rot);

	/*
	matvec(v2, rot, v);
	printf("Check");showvec(v2);
	*/

	// rotate about y-axis so the eye is on the z-axis
	rot[2][1] = rot[1][2] = 0;
	rot[1][1] = 1;
	g = sqrt(f * f + v[0] * v[0]);
	rot[0][0] = rot[2][2] = -f / g;
	rot[0][2] = -(rot[2][0] = v[0] / g);

	matmat(combo2, rot, combo);

	/*
	matvec(v2, combo2, v);
	printf("Check");showvec(v2);
	*/

	matvec(v, combo2, s.up);

	// rotate about z-axis so s.up is up
	// actually it points down, but the screen coords
	// are upside-down so it cancels out
	rot[0][2] = rot[2][0] = 0;
	rot[2][2] = 1;
	f = sqrt(v[0] * v[0] + v[1] * v[1]);
	rot[0][0] = rot[1][1] = -v[1] / f;
	rot[1][0] = -(rot[0][1] = v[0] / f);

	matmat(combo, rot, combo2);

	/*
	matvec(v2, combo, s.up);
	printf("Check");showvec(v2);
	matvec(v2, combo, s.eye);
	printf("Check");showvec(v2);
	*/

	addvec(v, s.eye, tr);
	matvec(v2, combo, v);
	printf("Eye now at");
	showvec(v2);
	printf("\n");
	seye = v2[2];

	/* Not accurate enough for this:
	if (v2[0] != 0.0 || v2[1] != 0.0) {
		fprintf(stderr, "Transformations didn't work!\n");
		showvec(v2);
		exit(1);
	}
	*/

	printf("New vectors are:\n");
	for (i=0; i<vcount; i++) {
		addvec(v, vertex[i], tr);
		matvec(svertex[i], combo, v);
		showvec(svertex[i]);
	}

	for (i=0; i<objcount; i++) {
		for (j=0; j<o[i].fcount; j++) {
			// Don't translate normals
			matvec(o[i].face[j].sn, combo, o[i].face[j].n);
			subvec(v, svertex[o[i].face[j].e[0][1]], svertex[o[i].face[j].e[0][0]]);
			printf("zero: %lf\n", dotvec(v, o[i].face[j].sn));
		}
	}

}

// Convert to macros for speed? Or will gcc automatically inline them?
void addvec(double *a, double *b, double *c) {
	a[0] = b[0] + c[0];
	a[1] = b[1] + c[1];
	a[2] = b[2] + c[2];
}

void subvec(double *a, double *b, double *c) {
	a[0] = b[0] - c[0];
	a[1] = b[1] - c[1];
	a[2] = b[2] - c[2];
}

void multvec(double a, double *v) {
	v[0] *= a;
	v[1] *= a;
	v[2] *= a;
}

double dotvec(double *a, double *b) {
	return(a[0]*b[0] + a[1]*b[1] + a[2]*b[2]);
}

// I hope gcc unrolls these
void matvec(double *r, double a[3][3], double *v) {
	int i, j;

	for (i=0; i<3; i++) {
		r[i] = 0;
		for (j=0; j<3; j++) {
			r[i] += a[i][j] * v[j];
		}
	}
}

void matmat(double a[3][3], double b[3][3], double c[3][3]) {
	int i, j, k;

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			a[i][j] = 0;
			for (k=0; k<3; k++) a[i][j] += b[i][k] * c[k][j];
		}
	}
}

void showvec(double *a) {
	int i;

	for (i=0; i<3; i++) {
		printf(" %lf", a[i]);
	}
	printf("\n");
}

void showmat(double a[3][3]) {
	int i;

	for (i=0; i<3; i++) {
		showvec(a[i]);
	}
}
